/***************************************************************************
 Example code to read keys from the Layad Circuits Alphanumeric Keyboard Module.
 Upload this code to an Arduino.
 Connect the TX, RX, and GND pins of the Arduino to the corresponding pins of the LC Alphanumeric Keyboard Module.
  
 This software is free provided that this notice is not removed and proper attribution 
 is accorded to Layad Circuits and its Author(s).
 Layad Circuits invests resources in producing free software. By purchasing Layad Circuits'
 products or utilizing its services, you support the continuing development of free software
 for all.
  
 Author(s): D. Deponio for Layad Circuits Electronics Engineering
 Revision: 1.0 - 2019/09/19 - initial creation
 Layad Circuits Electronics Engineering Supplies and Services
 B314 Lopez Bldg., Session Rd. cor. Assumption Rd., Baguio City, Philippines
 www.layadcircuits.com
 general: info@layadcircuits.com
 sales: sales@layadcircuits.com
 +63-916-442-8565
 ***************************************************************************/


/***************************************************************************
Example code to demonstrate the use of an alphanumeric keyboard module
and a 16x2 I2C character LCD module to input data to the ESP32.

MIT License

Copyright (c) [2020] lccam2020@gmail.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 ***************************************************************************/

#include <LiquidCrystal_I2C.h>
#include "WiFi.h"
#include "EEPROM.h"

#define VERSION_SHOWCREDS // Uncomment this if you want Wi-Fi credentials with F7 and F8.

#define TRM_PORT Serial // Uncomment to enable terminal/debug port.
#ifdef TRM_PORT
#define TRM_BAUD 9600
#define TRM_MAIN_SM
#define TRM_WIFICONN_SM
#endif

#define LCKB_PORT Serial1

enum LCKey{
  LCKey_up = 0x11,
  LCKey_left,//0x12
  LCKey_down,//0x13
  LCKey_right,//0x14
  LCKey_F1 = 0x82,
  LCKey_F2, //0x83
  LCKey_F3, //0x84
  LCKey_F4, //0x85
  LCKey_F5, //0x86
  LCKey_F6, //0x87
  LCKey_F7, //0x88
  LCKey_F8, //0x89
  LCKey_F9 //0x8A
};

// Pin definitions.
const uint8_t PIN_OU_LED_HB = 2;
const uint8_t PIN_KB_RX = 18;
const uint8_t PIN_KB_TX = 19;

// Minimum and maximum number of characters for SSID and password.
const uint8_t SSID_QTY_MINLEN = 5;
const uint8_t SSID_QTY_MAXLEN = 64;
const uint8_t PASSWORD_QTY_MINLEN = 5;
const uint8_t PASSWORD_QTY_MAXLEN = 64;
const uint8_t IP_QTY_MINLEN = 7; // 0.0.0.0
const uint8_t IP_QTY_MAXLEN = 15;

// Save locations in non volatile memory.
const uint16_t EEPROM_OFFSET_APSSID = 0x0000;
const uint16_t EEPROM_OFFSET_APPASS = EEPROM_OFFSET_APSSID + SSID_QTY_MAXLEN+1;
const uint16_t EEPROM_OFFSET_IPADD = EEPROM_OFFSET_APPASS + PASSWORD_QTY_MAXLEN+1;

const uint8_t LCD_QTY_MAXCOL = 16;
const uint8_t LCD_QTY_MAXROW = 2;

const uint32_t MENU_TSPAN_KBINPUTTTOUT = 10000; // Menu timesout if no activity after MENU_TSPAN_KBINPUTTTOUT.

LiquidCrystal_I2C lcd(0x27,LCD_QTY_MAXCOL,LCD_QTY_MAXROW);
// LiquidCrystal_I2C lcd(0x3E,LCD_QTY_MAXCOL,LCD_QTY_MAXROW);

bool gIsLcdFree;
bool gIsConnectedToRtr;
bool gIsCredNew;

char gKey;
bool gKeyIsAvailable{};

// ========== Helper functions start ==========
bool hlpr_isAlphaNum(char key){
  bool retVal = false;
  if(key >= 'A' && key <= 'Z') retVal = true;
  else if(key >= 'a' && key <= 'z') retVal = true;
  else if(key >= '0' && key <= '9') retVal = true;
  return retVal;
}

// Returns the length of a string, counting only printable characters.
// Counting stops if a non-printable character is found.
uint16_t hlpr_strlenPr(char *ptr){
 uint16_t ctr = 0;
 while(isprint(*(ptr+ctr))){ ctr++; }
 return ctr;
}

// Returns true if a string has the correct number of printable characters.
bool hlpr_checkStrlenPrintable(char *ptr, uint8_t minLen, uint8_t maxLen){
  uint16_t len = hlpr_strlenPr(ptr);
  return (len >= minLen && len <= maxLen);
}

// Returns true if a valid SSID is found in EEPROM. If so, SSID will be copied to ssidDest.
bool hlpr_getSsid(char * ssidDest){
  char ssidBfr[SSID_QTY_MAXLEN+1]{};
  String ssidStr = EEPROM.readString(EEPROM_OFFSET_APSSID);
  ssidStr.toCharArray(ssidBfr, SSID_QTY_MAXLEN);
  bool isLenOk = hlpr_checkStrlenPrintable(ssidBfr, SSID_QTY_MINLEN, SSID_QTY_MAXLEN);
  if(isLenOk) memcpy(ssidDest, ssidBfr, SSID_QTY_MAXLEN);
  return isLenOk;
}

void hlpr_setSsid(char * ssidSrc){
  char ssidBfr[SSID_QTY_MAXLEN+1]{};
  memcpy(ssidBfr, ssidSrc, SSID_QTY_MAXLEN);
  EEPROM.writeString(EEPROM_OFFSET_APSSID, ssidBfr);
  EEPROM.commit();
}

// Returns true if a valid PASSWORD is found in EEPROM. If so, PASSWORD will be copied to passwordDest.
bool hlpr_getPassword(char * passwordDest){
  char passwordBfr[PASSWORD_QTY_MAXLEN+1]{};
  String passwordStr = EEPROM.readString(EEPROM_OFFSET_APPASS);
  passwordStr.toCharArray(passwordBfr, PASSWORD_QTY_MAXLEN);
  bool isLenOk = hlpr_checkStrlenPrintable(passwordBfr, PASSWORD_QTY_MINLEN, PASSWORD_QTY_MAXLEN);
  if(isLenOk) memcpy(passwordDest, passwordBfr, PASSWORD_QTY_MAXLEN);
  return isLenOk;
}

void hlpr_setPassword(char * passwordSrc){
  char passwordBfr[PASSWORD_QTY_MAXLEN+1]{};
  memcpy(passwordBfr, passwordSrc, PASSWORD_QTY_MAXLEN);
  EEPROM.writeString(EEPROM_OFFSET_APPASS, passwordBfr);
  EEPROM.commit();
}

// Returns true if a valid IP address is found in EEPROM. If so, IP address will be copied to ipAddressDest.
bool hlpr_getIpAdd(char * ipAddressDest){
  char ipAddressBfr[IP_QTY_MAXLEN+2]{};
  String ipAddressStr = EEPROM.readString(EEPROM_OFFSET_IPADD);
  ipAddressStr.toCharArray(ipAddressBfr, IP_QTY_MAXLEN+1);
  bool isLenOk = hlpr_checkStrlenPrintable(ipAddressBfr, IP_QTY_MINLEN, IP_QTY_MAXLEN);
  if(isLenOk) memcpy(ipAddressDest, ipAddressBfr, IP_QTY_MAXLEN);
  return isLenOk;
}

void hlpr_setIpAdd(char * ipAddressSrc){
  char ipAddressBfr[IP_QTY_MAXLEN+1]{};
  memcpy(ipAddressBfr, ipAddressSrc, IP_QTY_MAXLEN);
  EEPROM.writeString(EEPROM_OFFSET_IPADD, ipAddressBfr);
  EEPROM.commit();
}

void hlpr_eraseEeprom(){
  for(uint16_t i=EEPROM_OFFSET_APSSID; i<(EEPROM_OFFSET_IPADD + IP_QTY_MAXLEN + 1); i++){
    EEPROM.writeByte(i, 0x00);
    EEPROM.commit();
  }
}

void hlpr_getMyIpAdd(char * ipAddressDest){
  char ipAddressBfr[IP_QTY_MAXLEN+2]{};
  String ipAddressStr = WiFi.localIP().toString();
  ipAddressStr.toCharArray(ipAddressBfr, IP_QTY_MAXLEN+1);
  memcpy(ipAddressDest, ipAddressBfr, IP_QTY_MAXLEN);
}

void ou_hb(){
  static uint32_t tRef;
  if(millis() - tRef < 500) return;
  digitalWrite(PIN_OU_LED_HB, !digitalRead(PIN_OU_LED_HB));

  tRef = millis();
}

// ========== Helper functions end   ==========

// This function maintains Wi-Fi connectivity.
void pr_wifiConn_sm(){
  static uint32_t tRef_lcd, tRef_conn;
  enum class St:uint8_t{
    init,
    checkCreds,
    connecting,
    connected,
    invalid
  };

  static St st;
  switch (st)
  {
    case St::init:
    {
      st = St::checkCreds;
      #ifdef TRM_WIFICONN_SM
      TRM_PORT.println(F("TRM_WIFICONN_SM: init > checkCreds!"));
      #endif
    }
      break;
    case St::checkCreds:
    {
      gIsConnectedToRtr = false;
      WiFi.disconnect(true, true);
      while (WiFi.status() == WL_CONNECTED); // NB: Blocking!!!
      char ssidBfr[SSID_QTY_MAXLEN+1]{};
      char passBfr[PASSWORD_QTY_MAXLEN+1]{};
      bool isSsidOk = hlpr_getSsid(ssidBfr);
      bool isPassOk = hlpr_getPassword(passBfr);
      if(isSsidOk && isPassOk)
      {
        tRef_lcd = millis();
        tRef_conn = millis();
        WiFi.begin(ssidBfr, passBfr);
        st = St::connecting;
        #ifdef TRM_WIFICONN_SM
        TRM_PORT.println(F("TRM_WIFICONN_SM: checkCreds > connecting!"));
        #endif
      }
      else{
        tRef_lcd = millis();
        st = St::invalid;
        #ifdef TRM_WIFICONN_SM
        TRM_PORT.println(F("TRM_WIFICONN_SM: checkCreds > invalid!"));
        #endif
      }
    }
      break;
    case St::connecting: // Connecting to AP.
    {
      static uint32_t tRef_connecting;
      static uint8_t st_connecting;
      switch (st_connecting)
      {
        case 0: // Connecting.
        {
          if(WiFi.status() == WL_CONNECTED){
            // gIsConnectedToRtr = true;
            if(gIsLcdFree){
              ///////////////////////////////////0123456789ABCDEF
              lcd.setCursor(0,1); lcd.print(F("ROUTR CONNECT OK"));
            }
            tRef_connecting = millis();
            st_connecting = 1;
          }
          else if(millis() - tRef_conn > 10000){
            WiFi.disconnect(true, true);
            while (WiFi.status() == WL_CONNECTED); // NB: Blocking!!!
            st = St::checkCreds;
            #ifdef TRM_WIFICONN_SM
            TRM_PORT.println(F("TRM_WIFICONN_SM: connecting > checkCreds, timeout"));
            #endif
          }
          else if(millis() - tRef_lcd > 500){
            if(gIsLcdFree){
              ///////////////////////////////////0123456789ABCDEF
              lcd.setCursor(0,1); lcd.print(F("CONECTNG TO ROTR"));
            }
            tRef_lcd = millis();
          }
        }
          break;
        case 1: // Delay.
        {
          if(millis() - tRef_connecting > 1000){
            if(gIsLcdFree){
              // lcd.clear();
              ///////////////////////////////////0123456789ABCDEF
              // lcd.setCursor(0,1); lcd.print(F("CONNECTION OK!  "));
            }
            gIsConnectedToRtr = true;
            st_connecting = 0;
            st = St::connected;
            #ifdef TRM_WIFICONN_SM
            TRM_PORT.println(F("TRM_WIFICONN_SM: connecting > connected!"));
            #endif
          }
        }
          break;
        default:
          break;
      }

    }
      break;
    case St::connected:
    {
      if(gIsCredNew){
        gIsCredNew = false;
        st = St::checkCreds;
        #ifdef TRM_WIFICONN_SM
        TRM_PORT.println(F("TRM_WIFICONN_SM: connected > checkCreds, gIsCredNew"));
        #endif
      }
      else if(WiFi.status() != WL_CONNECTED){
        st = St::checkCreds;
        #ifdef TRM_WIFICONN_SM
        TRM_PORT.println(F("TRM_WIFICONN_SM: connected > checkCreds, disconnected from router!"));
        #endif
      }
    }
      break;
    case St::invalid:
    {
      if(gIsCredNew){
        gIsCredNew = false;
        st = St::checkCreds;
        #ifdef TRM_WIFICONN_SM
        TRM_PORT.println(F("TRM_WIFICONN_SM: invalid > checkCreds, gIsCredNew"));
        #endif
      }
      else if(millis() - tRef_lcd > 500){
        if(gIsLcdFree){
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("INVALD RTR CREDS"));
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,1); lcd.print(F("F2:SSID F3:PASSW"));
        }
        tRef_lcd = millis();
      }
    }
      break;
    default:
      break;
  }
}

// This function runs the menu.
void pr_main_sm(){
  static uint32_t tRef;
  enum class St:uint8_t{
    init,
    clearLcd,
    idle,
    getIp,
    getApSsid,
    getApPw,
    delay
  };
  static St st;
  switch (st)
  {
    case St::init:
    {
      st = St::clearLcd;
      #ifdef TRM_MAIN_SM
      TRM_PORT.println(F("TRM_MAIN_SM: init > clearLcd!"));
      #endif
    }
      break;
    case St::clearLcd:
    {
      while(LCKB_PORT.available()) LCKB_PORT.read();
      if(gKeyIsAvailable) gKeyIsAvailable = false;
      lcd.clear();
      ///////////////////////////////////0123456789ABCDEF
      lcd.setCursor(0,0); lcd.print(F("ESP32 LAYAD CKTS"));
      lcd.setCursor(0,1); lcd.print(F("ALPHANUM KEYBRD "));
      gIsLcdFree = true;
      st = St::idle;
      #ifdef TRM_MAIN_SM
      TRM_PORT.println(F("TRM_MAIN_SM: clearLcd > idle!"));
      #endif
    }
      break;
    case St::idle:
    {
      if(gKeyIsAvailable){
        char key = gKey;
        gKeyIsAvailable = false;
        gIsLcdFree = false;
        if(key == LCKey_F2){
          tRef = millis();
          st = St::getApSsid;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > getApSsid!"));
          #endif
        }
        else if(key == LCKey_F3){
          tRef = millis();
          st = St::getApPw;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > getApPw!"));
          #endif
        }
        else if(key == LCKey_F5){
          char ipAddBfr[IP_QTY_MAXLEN+1]{};
          hlpr_getMyIpAdd(ipAddBfr);
          char lineBfr[LCD_QTY_MAXCOL + 1]{};
          memcpy(lineBfr, ipAddBfr, LCD_QTY_MAXCOL);
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("MY IP ADDR:     "));
          lcd.setCursor(0,1); lcd.print(lineBfr);
          tRef = millis();
          st = St::delay;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > delay!"));
          #endif
        }
        #ifdef VERSION_SHOWCREDS
        else if(key == LCKey_F7){
          char ssidBfr[SSID_QTY_MAXLEN+1]{};
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("ROUTER SSID:    "));
          if(hlpr_getSsid(ssidBfr)){
            char lineBfr[LCD_QTY_MAXCOL + 1]{};
            memcpy(lineBfr, ssidBfr, LCD_QTY_MAXCOL);
            lcd.setCursor(0,1); lcd.print(lineBfr);
          }
          else{ ////////////////////////////////0123456789ABCDEF
            lcd.setCursor(0,1); lcd.print(F("INVALID          "));
          }
          tRef = millis();
          st = St::delay;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > delay!"));
          #endif
        }
        else if(key == LCKey_F8){
          char passBfr[PASSWORD_QTY_MAXLEN+1]{};
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("ROUTER PASSWORD:"));
          if(hlpr_getPassword(passBfr)){
            char lineBfr[LCD_QTY_MAXCOL + 1]{};
            memcpy(lineBfr, passBfr, LCD_QTY_MAXCOL);
            lcd.setCursor(0,1); lcd.print(lineBfr);
          }
          else{ ////////////////////////////////0123456789ABCDEF
            lcd.setCursor(0,1); lcd.print(F("INVALID          "));
          }
          tRef = millis();
          st = St::delay;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > delay!"));
          #endif
        }
        #endif
        else if(key == LCKey_F9){
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("CLEAR WIFI CREDS"));
          hlpr_eraseEeprom();
          gIsCredNew = true;
          tRef = millis();
          st = St::delay;
          #ifdef TRM_MAIN_SM
          TRM_PORT.println(F("TRM_MAIN_SM: idle > delay!"));
          #endif
        }
        // else if (key == 'a'){
        //   // Process other keys here.
        // }
      }
    }
      break;
    case St::getApSsid:
    {
      static uint8_t keyCtr;
      static char inputBfr[SSID_QTY_MAXLEN+1];
      static uint8_t st_getApSsid;
      switch(st_getApSsid){
        case 0: // Initial LCD print.
        {
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("INPUT ROTR SSID:"));
          keyCtr = 0;
          memset(inputBfr, 0, sizeof(inputBfr));
          st_getApSsid = 1;
        }
          break;
        case 1: // Process input.
        {
          if(gKeyIsAvailable){
            tRef = millis();
            char key = gKey;
            gKeyIsAvailable = false;
            if(isprint(key)){
              if(keyCtr < SSID_QTY_MAXLEN){
                inputBfr[keyCtr++] = key;
                if(keyCtr <= LCD_QTY_MAXCOL){
                  lcd.setCursor(0, 1); lcd.print(inputBfr);
                }
                else{
                  char windowBfr[LCD_QTY_MAXCOL]{};
                  memcpy(windowBfr, &inputBfr[keyCtr - LCD_QTY_MAXCOL], LCD_QTY_MAXCOL);
                  lcd.setCursor(0, 1); lcd.print(windowBfr);
                }
              }
            }
            else if(key == '\n'){ // Enter.
              if(keyCtr >= SSID_QTY_MINLEN && keyCtr <= SSID_QTY_MAXLEN){
                ///////////////////////////////////0123456789ABCDEF
                lcd.setCursor(0,0); lcd.print(F("INPUT SSID: OK  "));
                inputBfr[keyCtr] = 0x00;
                hlpr_setSsid(inputBfr);
                st_getApSsid = 0;
                gIsCredNew = true;
                st = St::delay;
                #ifdef TRM_MAIN_SM
                TRM_PORT.println(F("TRM_MAIN_SM: getApSsid > delay, ENTER"));
                TRM_PORT.println(inputBfr);
                #endif
              }
            }
            else if(key == '\b' ){ // Backspace.
              if(keyCtr > 0){
                inputBfr[--keyCtr] = ' ';
                // lcd.setCursor(0, 1); lcd.print(inputBfr);
                if(keyCtr <= LCD_QTY_MAXCOL){
                  lcd.setCursor(0, 1); lcd.print(inputBfr);
                }
                else{
                  char windowBfr[LCD_QTY_MAXCOL]{};
                  memcpy(windowBfr, &inputBfr[keyCtr - LCD_QTY_MAXCOL], LCD_QTY_MAXCOL);
                  lcd.setCursor(0, 1); lcd.print(windowBfr);
                }
              }
            }
            else if(key == '\e' ){ // Escape.
              st_getApSsid = 0;
              st = St::clearLcd;
              #ifdef TRM_MAIN_SM
              TRM_PORT.println(F("TRM_MAIN_SM: getApSsid > clearLcd, ESC"));
              #endif
            }
          }
          else if(millis() - tRef > MENU_TSPAN_KBINPUTTTOUT){
            st_getApSsid = 0;
            st = St::clearLcd;
            #ifdef TRM_MAIN_SM
            TRM_PORT.println(F("TRM_MAIN_SM: getApSsid > clearLcd"));
            #endif
          }
        }
          break;
        default:
          break;
      }
    }
      break;
    case St::getApPw:
    {
      static uint8_t keyCtr;
      static char inputBfr[PASSWORD_QTY_MAXLEN+1];
      static uint8_t st_getApPw;
      switch(st_getApPw){
        case 0: // Initial LCD print.
        {
          lcd.clear();
          ///////////////////////////////////0123456789ABCDEF
          lcd.setCursor(0,0); lcd.print(F("INPUT PASSWORD: "));
          keyCtr = 0;
          memset(inputBfr, 0, sizeof(inputBfr));
          st_getApPw = 1;
        }
          break;
        case 1: // Process input.
        {
          if(gKeyIsAvailable){
            tRef = millis();
            char key = gKey;
            gKeyIsAvailable = false;
            if(isprint(key)){
              if(keyCtr < PASSWORD_QTY_MAXLEN){
                inputBfr[keyCtr++] = key;
                if(keyCtr <= LCD_QTY_MAXCOL){
                  lcd.setCursor(0, 1); lcd.print(inputBfr);
                }
                else{
                  char windowBfr[LCD_QTY_MAXCOL]{};
                  memcpy(windowBfr, &inputBfr[keyCtr - LCD_QTY_MAXCOL], LCD_QTY_MAXCOL);
                  lcd.setCursor(0, 1); lcd.print(windowBfr);
                }
              }
            }
            else if(key == '\n'){ // Enter.
              if(keyCtr >= PASSWORD_QTY_MINLEN && keyCtr <= PASSWORD_QTY_MAXLEN){
                ///////////////////////////////////0123456789ABCDEF
                lcd.setCursor(0,0); lcd.print(F("INPUT PASSWRD:OK"));
                inputBfr[keyCtr] = 0x00;
                hlpr_setPassword(inputBfr);
                gIsCredNew = true;
                st_getApPw = 0;
                st = St::delay;
                #ifdef TRM_MAIN_SM
                TRM_PORT.println(F("TRM_MAIN_SM: getApPw > delay, ENTER"));
                TRM_PORT.println(inputBfr);
                #endif
              }
            }
            else if(key == '\b' ){ // Backspace.
              if(keyCtr > 0){
                inputBfr[--keyCtr] = ' ';
                // lcd.setCursor(0, 1); lcd.print(inputBfr);
                if(keyCtr <= LCD_QTY_MAXCOL){
                  lcd.setCursor(0, 1); lcd.print(inputBfr);
                }
                else{
                  char windowBfr[LCD_QTY_MAXCOL]{};
                  memcpy(windowBfr, &inputBfr[keyCtr - LCD_QTY_MAXCOL], LCD_QTY_MAXCOL);
                  lcd.setCursor(0, 1); lcd.print(windowBfr);
                }
              }
            }
            else if(key == '\e' ){ // Escape.
              st_getApPw = 0;
              st = St::clearLcd;
              #ifdef TRM_MAIN_SM
              TRM_PORT.println(F("TRM_MAIN_SM: getApPw > clearLcd, ESC"));
              #endif
            }
          }
          else if(millis() - tRef > MENU_TSPAN_KBINPUTTTOUT){
            st_getApPw = 0;
            st = St::clearLcd;
            #ifdef TRM_MAIN_SM
            TRM_PORT.println(F("TRM_MAIN_SM: getApPw > clearLcd"));
            #endif
          }
        }
          break;
        default:
          break;
      }
    }
      break;
    case St::delay:
    {
      if(gKeyIsAvailable){
        gKeyIsAvailable = false;;
        st = St::clearLcd;
        #ifdef TRM_MAIN_SM
        TRM_PORT.println(F("TRM_MAIN_SM: delay > clearLcd, key"));
        #endif
      }
      else if(millis() - tRef > 3000){
        st = St::clearLcd;
        #ifdef TRM_MAIN_SM
        TRM_PORT.println(F("TRM_MAIN_SM: delay > clearLcd, wait done"));
        #endif
      }
    }
      break;
    default:
      break;
  }
}

// This function buffers a key from the alphanumeric keyboard module.
void in_lcKb(){
  if(LCKB_PORT.available()){
    gKey = LCKB_PORT.read();
    gKeyIsAvailable = true;
  }
}

void setup() {
 // put your setup code here, to run once:
  // Init heartbeat LED
  pinMode(PIN_OU_LED_HB, OUTPUT);
  digitalWrite(PIN_OU_LED_HB, LOW);

  #ifdef TRM_PORT
  TRM_PORT.begin(TRM_BAUD);
  TRM_PORT.println(F("ESP32 Layad Circuits Alphanumeric Keyboard Demo"));
  #endif

  // Init LCD
  lcd.init(); 
  lcd.backlight();
  lcd.clear();
  //////////////////////////////////0123456789ABCDEF
  lcd.setCursor(0,0); lcd.print(F("ESP32 LAYAD CKTS"));
  lcd.setCursor(0,1); lcd.print(F("ALPHANUM KEYBRD "));

  // LC Alphanumeric Keyboard
  LCKB_PORT.begin(9600, SERIAL_8N1, PIN_KB_RX, PIN_KB_TX);

  // Non-volatile memory
  if (!EEPROM.begin(1000)) {
    /////////////////////////////////0123456789ABCDEF
    lcd.setCursor(0,1); lcd.print(F("DATABASE FAILED!"));
    #ifdef TRM_PORT
    TRM_PORT.println("Failed to initialise EEPROM!");
    TRM_PORT.println("Restarting...");
    #endif
    delay(1000);
    ESP.restart();
  }
}

void loop() {
 // put your main code here, to run repeatedly:
  in_lcKb();
  pr_wifiConn_sm();
  pr_main_sm();
  ou_hb();
}
